#include <cstdio>

using std::printf;
using std::putchar;
using std::fopen;
using std::fread;
using std::fclose;

#include "rom.hh"
#include "romaddr.hh"

std::string InputFN = "megaman1.nes";
std::string OutputFN = "tmp.nes";

class ROM ROM;

ROM::ROM()
{
    ReRead();
}

void ROM::ReRead()
{
    FILE*fp = fopen(InputFN.c_str(), "rb");
    if(!fp) { perror(InputFN.c_str()); return; }
    
    char header[0x10];
    fread(header,1,0x10, fp);
    
    data.resize((ROMmap_npages = header[4]) * 0x4000);
    
    fread(&data[0], 1, data.size(), fp);
    fclose(fp);
}

void ROM::Resize(unsigned npages)
{
    unsigned oldcap = data.size();
    unsigned cap    = npages * 0x4000;
    
    if(cap > oldcap)
    {
        unsigned spacebegin  = oldcap-0x4000;
        unsigned spaceamount = cap-oldcap;
        while(spaceamount >= 0x4000)
        {
            space.Add(ROM2Space(spacebegin), 0x4000);
            spaceamount -= 0x4000;
            spacebegin  += 0x4000;
        }
        data.resize(cap);
    }
    
    if(oldcap != cap && cap >= 0x4000 && oldcap >= 0x4000)
    {
        std::memcpy(&data[cap-0x4000],
                    &data[oldcap-0x4000],
                     0x4000);
        std::memset(&data[oldcap-0x4000], 0, 0x4000);
    }

    if(cap < oldcap)
    {
        unsigned spacebegin  = cap-0x4000;
        unsigned spaceamount = oldcap-cap;
        while(spaceamount >= 0x4000)
        {
            space.Del(ROM2Space(spacebegin), 0x4000);
            spaceamount -= 0x4000;
            spacebegin  += 0x4000;
        }
        data.resize(cap);
    }

    ROMmap_npages = npages;
}

ROM::~ROM()
{
    FILE*fp = fopen(OutputFN.c_str(), "wb");
    if(!fp) { perror(OutputFN.c_str()); return; }
    
    const std::vector<unsigned> o65addrs = O65linker::GetAddrList();
    for(unsigned a=0; a<o65addrs.size(); ++a)
    {
        unsigned addr = o65addrs[a];
        if(addr == NOWHERE) continue;
        
        const std::vector<unsigned char>& code = O65linker::GetCode(a);
        const std::string& name = O65linker::GetName(a);
        
        std::fprintf(stderr, "Writing '%s' (%u bytes) to $%X ($%X in ROM)\n",
            name.c_str(), code.size(), addr, NES2ROMaddr(addr));
        WriteBytes(addr, code.size(), &code[0]);
    }
    
    char Header[16] = {'N','E','S',26, ROMmap_npages,0,'!',0,
                       0,0,0,0,0,0,0,0};
    fwrite(Header,1,16,fp);
    fwrite(&data[0], 1, data.size(), fp);
    fclose(fp);
}

#include "msginsert.hh"

void MessageLinkingModules(unsigned count)
{
    //printf("Linking %u modules\n", count);
}
 
void MessageLoadingItem(const std::string& header)
{
    //printf("Loading %s\n", header.c_str());
}

void MessageWorking()
{
}

void MessageDone()
{
    //printf("- done\n");
}

void MessageModuleWithoutAddress(const std::string& name)
{
    fprintf(stderr, "O65 linker: Module %s is still without address\n", name.c_str());
}
 
void MessageUndefinedSymbol(const std::string& name)
{
    fprintf(stderr, "O65 linker: Symbol '%s' still undefined\n", name.c_str());
}

void MessageDuplicateDefinition(const std::string& name, unsigned nmods, unsigned ndefs)
{
    fprintf(stderr, "O65 linker: Symbol '%s' defined in %u module(s) and %u global(s)\n",
        name.c_str(), nmods, ndefs);
}

void MessageUndefinedSymbols(unsigned n)
{
    fprintf(stderr, "O65 linker: Still %u undefined symbol(s)\n", n);
}
